from collections import OrderedDict
from typing import Any, Literal, Optional, TypedDict, Union

import networkx as nx
import yaml
from nomad.utils import get_logger
from pydantic import BaseModel, Field

logger = get_logger(__name__)
TASK_M_DEF = 'nomad.datamodel.metainfo.workflow.TaskReference'
WORKFLOW_M_DEF = 'nomad.datamodel.metainfo.workflow.TaskReference'
# TODO not yet sure about the specification of actual tasks, need to test

SectionType = Literal['task', 'workflow', 'input', 'output', 'other']
EntryType = Literal['simulation']
# TODO check/implement functionality of "other" type


# Define a custom representer for OrderedDict
def represent_ordereddict(dumper, data):
    return dumper.represent_dict(data.items())


class SingleQuotedScalarString(str):
    pass


def single_quoted_scalar_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style="'")


yaml.add_representer(SingleQuotedScalarString, single_quoted_scalar_representer)
yaml.add_representer(str, single_quoted_scalar_representer)

# Register the custom representer
yaml.add_representer(OrderedDict, represent_ordereddict)


class PathInfo(TypedDict, total=False):
    upload_id: str
    entry_id: str
    mainfile_path: str
    supersection_path: str
    supersection_index: int
    section_type: str
    section_index: int
    archive_path: str


default_path_info = {
    'upload_id': None,
    'entry_id': None,
    'mainfile_path': '',
    'supersection_path': '',
    'supersection_index': None,
    'section_type': None,
    'section_index': -1,
    'archive_path': '',
}


class NomadSection(BaseModel):
    name: Optional[str] = Field(None, description='Name of the section')
    type: Optional[SectionType] = Field(None, description='Type of the section')
    path_info: dict[str, Any] = Field(
        default=default_path_info.copy(), description='Archive path'
    )
    inputs: list[dict[str, Any]] = Field(
        [{}],
        description='section inputs',
    )
    outputs: list[dict[str, Any]] = Field([{}], description='section outputs')

    def __init__(self, **data):
        super().__init__(**data)
        self.path_info = {**default_path_info, **self.path_info}

    @property
    def archive_path(self) -> str:
        if not self.path_info:
            logger.warning(
                f'No path info provided for {self.type}-{self.name}.'
                ' Section reference will be missing.'
            )
            return ''

        if self.path_info.get('archive_path'):
            return self.path_info['archive_path']
        elif self.type == 'workflow':
            return 'workflow2'
        else:
            return self._get_supersection_path()

    def _get_supersection_path(self) -> str:
        archive_path = ''
        if self.path_info.get('supersection_path'):
            archive_path = self._get_supersection_path_with_index()
        elif self.path_info.get('section_type'):
            archive_path = self._get_section_type_path()
        else:
            logger.warning(
                (
                    'No supersection path or section type provided for '
                    f'{self.type}-{self.name}. Section reference may be incorrect.'
                ),
            )
        return self._get_section_path(archive_path)

    def _get_supersection_path_with_index(self) -> str:
        archive_path = self.path_info['supersection_path']
        if self.path_info.get('supersection_index') is not None:
            archive_path += f"/{self.path_info.get('supersection_index')}"
        return archive_path

    def _get_section_type_path(self) -> str:
        archive_path = ''
        if self.path_info.get('section_type') in ['system', 'calculation', 'method']:
            run_index = self.path_info.get('supersection_index', 0)
            run_index = run_index if run_index is not None else 0
            archive_path = f'run/{run_index}'
        elif self.path_info.get('section_type') in ['results']:
            archive_path = 'workflow2'
        else:
            archive_path += f"/{self.path_info.get('section_type')}"
            if self.path_info.get('section_index') is not None:
                archive_path += f"/{self.path_info.get('section_index')}"
        return archive_path

    def _get_section_path(self, archive_path: str) -> str:
        if self.path_info.get('section_type') is not None:
            archive_path += f"/{self.path_info['section_type']}"
            if self.path_info.get('section_index') is not None:
                archive_path += f"/{self.path_info['section_index']}"
        else:
            # TODO: this may be is a false warning if the subsection_path is given
            logger.warning(
                (
                    f'No section type provided for {self.type}-{self.name}. '
                    'Section reference may be incorrect.'
                ),
            )
        return archive_path

    @property
    def upload_prefix(self) -> str:
        if not self.path_info['mainfile_path']:
            logger.warning(
                f'No mainfile path provided for {self.type}-{self.name}. '
                'Section reference will be missing.'
            )
            return ''

        if self.path_info.get('entry_id'):
            upload_prefix = f"/entries/{self.path_info.get('entry_id')}"
        elif self.path_info.get('upload_id'):
            upload_prefix = f"/uploads/{self.path_info.get('upload_id')}"
        else:
            upload_prefix = f"../upload{''}"

        return f"{upload_prefix}/archive/mainfile/{self.path_info['mainfile_path']}"

    @property
    def full_path(self) -> str:
        if not self.upload_prefix or not self.archive_path:
            return ''

        return f"{self.upload_prefix}#/{self.archive_path}{''}"

    def to_dict(self) -> dict:
        return OrderedDict(
            {'name': self.name, 'section': SingleQuotedScalarString(self.full_path)}
        )


class NomadTask(BaseModel):
    name: str
    m_def: str
    inputs: list[NomadSection] = Field(default_factory=list)
    outputs: list[NomadSection] = Field(default_factory=list)
    task_section: Optional[NomadSection] = None

    def __init__(self, **data):
        super().__init__(**data)
        for i, input_ in enumerate(self.inputs):
            if input_.name is None:
                input_.name = f'input_{i}'
        for o, output_ in enumerate(self.outputs):
            if output_.name is None:
                output_.name = f'output_{o}'

    @property
    def m_def(self) -> str:
        if self.task_section.type == 'workflow':
            return WORKFLOW_M_DEF
        elif self.task_section.type == 'task':
            return TASK_M_DEF

    @property
    def task(self) -> Optional[str]:
        if self.task_section.type == 'workflow' and self.task_section.upload_prefix:
            return self.task_section.upload_prefix + '#/workflow2'
        elif self.task_section.type == 'task' and self.task_section.full_path:
            return self.task_section.full_path
        else:
            return None

    def to_dict(self) -> dict:
        output_dict = OrderedDict()
        if self.m_def:
            output_dict['m_def'] = self.m_def
        output_dict['name'] = self.name
        if self.task:
            output_dict['task'] = self.task
        output_dict['inputs'] = [i.to_dict() for i in self.inputs]
        output_dict['outputs'] = [o.to_dict() for o in self.outputs]

        return output_dict


class NomadWorkflowArchive(BaseModel):
    archive_section: str = None
    name: str = None
    inputs: list[NomadSection] = Field(default_factory=list)
    outputs: list[NomadSection] = Field(default_factory=list)
    tasks: list[NomadTask] = Field(default_factory=list)

    def remove_duplicate_ios(self) -> None:
        def remove_duplicates(ios):
            seen = set()
            trimmed = []
            for io in ios:
                if io.full_path not in seen:
                    trimmed.append(io)
                    seen.add(io.full_path)
            return trimmed

        self.inputs = remove_duplicates(self.inputs)
        self.outputs = remove_duplicates(self.outputs)

    def to_dict(self) -> dict:
        yaml_dict = {self.archive_section: OrderedDict({})}
        if self.name:
            yaml_dict[self.archive_section]['name'] = self.name
        if self.inputs:
            yaml_dict[self.archive_section]['inputs'] = [
                i.to_dict() for i in self.inputs
            ]
        if self.outputs:
            yaml_dict[self.archive_section]['outputs'] = [
                o.to_dict() for o in self.outputs
            ]
        if self.tasks:
            yaml_dict[self.archive_section]['tasks'] = [t.to_dict() for t in self.tasks]

        return yaml_dict

    def to_yaml(self, destination_filename: str) -> None:
        with open(destination_filename, 'w') as f:
            yaml.dump(
                self.to_dict(),
                f,
                default_flow_style=False,
                allow_unicode=True,
                width=80,
            )


class NomadWorkflow(BaseModel):
    destination_filename: str
    archive_section: str
    name: str
    node_attributes: dict[int, Any] = {}
    workflow_graph: nx.DiGraph = None
    task_elements: dict[str, NomadSection] = Field(default_factory=dict)
    simulation_default_sections: dict[str, list[str]] = Field(
        default_factory=dict,
        description='Default input and output sections for simulation tasks',
    )

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)
        self.task_elements = {}
        self.simulation_default_sections = {
            'inputs': ['system'],
            'outputs': ['system', 'calculation'],
        }
        # ! add more defaults here
        if self.workflow_graph is None:
            self.workflow_graph = nodes_to_graph(self.node_attributes)
        self.fill_workflow_graph()

    def register_section(
        self, node_key: Union[int, str, tuple], node_attrs: dict[str, Any]
    ) -> None:
        section = NomadSection(**node_attrs)
        self.task_elements[node_key] = section

    def fill_workflow_graph(self) -> None:
        """_summary_"""
        for node_source, node_dest, edge in list(self.workflow_graph.edges(data=True)):
            self._resolve_edge_inputs(node_source, node_dest, edge)
            self._resolve_edge_outputs(node_source, node_dest, edge)
            self._add_defaults(node_source, node_dest, edge)

    def _resolve_edge_inputs(self, node_source, node_dest, edge) -> None:
        if not edge.get('inputs'):
            nx.set_edge_attributes(
                self.workflow_graph, {(node_source, node_dest): {'inputs': []}}
            )
        for input_ in edge['inputs']:
            if not input_.get('path_info', {}):
                continue
            if not input_['path_info'].get('mainfile_path', ''):
                input_['path_info']['mainfile_path'] = self._get_mainfile_path(
                    node_source
                )

    def _resolve_edge_outputs(self, node_source, node_dest, edge) -> None:
        if not edge.get('outputs'):
            nx.set_edge_attributes(
                self.workflow_graph, {(node_source, node_dest): {'outputs': []}}
            )
        for output_ in edge.get('outputs', []):
            if not output_.get('path_info', {}):
                continue
            if not output_['path_info'].get('mainfile_path', ''):
                node_source_type = self.workflow_graph.nodes[node_source].get(
                    'type', ''
                )
                if node_source_type == 'input':
                    output_['path_info']['mainfile_path'] = self._get_mainfile_path(
                        node_dest
                    )
                else:
                    output_['path_info']['mainfile_path'] = self._get_mainfile_path(
                        node_source
                    )

    def _add_defaults(self, node_source, node_dest, edge) -> None:
        if self.workflow_graph.nodes[node_source].get('type', '') in [
            'task',
            'workflow',
        ]:
            for outputs_ in self._get_defaults('outputs', node_source, node_dest):
                edge['inputs'].append(outputs_)
                # add the output to the graph
                self.workflow_graph.add_node(
                    len(self.workflow_graph.nodes), type='output', **outputs_
                )
                self.workflow_graph.add_edge(
                    node_source, len(self.workflow_graph.nodes) - 1
                )
        if self.workflow_graph.nodes[node_dest].get('type', '') in ['task', 'workflow']:
            for inputs_ in self._get_defaults('inputs', node_source, node_dest):
                edge['outputs'].append(inputs_)
                # add the input to the graph
                self.workflow_graph.add_node(
                    len(self.workflow_graph.nodes), type='input', **inputs_
                )
                self.workflow_graph.add_edge(
                    len(self.workflow_graph.nodes) - 1, node_dest
                )

    def _get_mainfile_path(self, node):
        return (
            self.workflow_graph.nodes[node]
            .get('path_info', '')
            .get('mainfile_path', '')
        )

    def _check_for_defaults(self, inout_type, default_section, edge) -> bool:
        inout_type = 'inputs' if inout_type == 'outputs' else 'outputs'
        for input_ in edge.get(inout_type, []):
            if input_.get('path_info', {}).get('section_type', '') == default_section:
                return True
            if (
                input_.get('path_info', {}).get('supersection_path', '')
                == default_section
            ):
                return True
        return False

    def _get_defaults(
        self, inout_type: Literal['inputs', 'outputs'], node_source, node_dest
    ) -> list:
        entry_type, partner_node = self._determine_entry_type_and_partner_node(
            node_source, node_dest
        )

        if entry_type == 'simulation':
            return self._get_simulation_defaults(
                inout_type, partner_node, node_source, node_dest
            )
        else:
            return self._get_general_defaults(
                inout_type, partner_node, node_source, node_dest
            )

    def _determine_entry_type_and_partner_node(self, node_source, node_dest):
        entry_type_source = self.workflow_graph.nodes[node_source].get('entry_type', '')
        node_source_type = self.workflow_graph.nodes[node_source].get('type', '')
        entry_type_dest = self.workflow_graph.nodes[node_dest].get('entry_type', '')
        node_dest_type = self.workflow_graph.nodes[node_dest].get('type', '')
        entry_type = ''
        partner_node = node_source
        if entry_type_source == 'simulation':
            entry_type = 'simulation'
        elif entry_type_dest == 'simulation' and node_source_type == 'input':
            entry_type = 'simulation'
            partner_node = node_dest
        elif node_dest_type == 'output':
            partner_node = node_dest
        return entry_type, partner_node

    def _get_general_defaults(self, inout_type, partner_node, node_source, node_dest):
        section = NomadSection(**self.workflow_graph.nodes[partner_node])
        archive_path = section.archive_path
        default_sections = {
            'inputs': [archive_path],
            'outputs': [archive_path],
        }

        inouts = []
        for default_section in default_sections[inout_type]:
            if not self._flag_defaults(
                inout_type, default_section, node_source, node_dest
            ):
                partner_name = self.workflow_graph.nodes[partner_node].get('name', '')
                path_info = self.workflow_graph.nodes[partner_node].get('path_info', {})
                path_info['supersection_path'] = default_section
                inouts.append(
                    {
                        'name': (
                            f'{inout_type[:-1]} {default_section} '
                            f'from {partner_name}'
                        ),
                        'path_info': path_info,
                        'is_default': True,
                    },
                )
        return inouts

    def _get_simulation_defaults(
        self, inout_type, partner_node, node_source, node_dest
    ):
        default_sections = self.simulation_default_sections
        inouts = []
        for default_section in default_sections[inout_type]:
            if not self._flag_defaults(
                inout_type, default_section, node_source, node_dest
            ):
                partner_name = self.workflow_graph.nodes[partner_node].get('name', '')
                inouts.append(
                    {
                        'name': (
                            f'{inout_type[:-1]} {default_section} '
                            f'from {partner_name}'
                        ),
                        'path_info': {
                            'section_type': default_section,
                            'mainfile_path': self._get_mainfile_path(partner_node),
                        },
                    },
                )
        return inouts

    def _flag_defaults(self, inout_type, default_section, node_source, node_dest):
        flag_defaults = False
        if inout_type == 'outputs':
            for _, _, edge2 in self.workflow_graph.out_edges(node_source, data=True):
                if self._check_for_defaults(inout_type, default_section, edge2):
                    flag_defaults = True
                    break
        elif inout_type == 'inputs':
            for _, _, edge2 in self.workflow_graph.in_edges(node_dest, data=True):
                if self._check_for_defaults(inout_type, default_section, edge2):
                    flag_defaults = True
                    break
        return flag_defaults

    #! Keep the below old implementation until full testing for ruff refactoring above!
    # def _get_defaults(
    #     self, inout_type: Literal['inputs', 'outputs'], node_source, node_dest
    # ) -> list:
    #     entry_type_source = self.workflow_graph.nodes[node_source].get('entry_type',
    # '')
    #     node_source_type = self.workflow_graph.nodes[node_source].get('type', '')
    #     entry_type_dest = self.workflow_graph.nodes[node_dest].get('entry_type', '')
    #     node_dest_type = self.workflow_graph.nodes[node_dest].get('type', '')
    #     entry_type = ''
    #     partner_node = node_source
    #     if entry_type_source == 'simulation':
    #         entry_type = 'simulation'
    #     elif entry_type_dest == 'simulation' and node_source_type == 'input':
    #         entry_type = 'simulation'
    #         partner_node = node_dest
    #     elif node_dest_type == 'ouput':
    #         partner_node = node_dest

    #     # TODO - unify this with the simulation defaults case
    #     # TODO - for sim case, fix prefix to full path by copying path_info,
    #     # TODO - like in general case
    #     default_sections = {}
    #     if not entry_type:
    #         section = NomadSection(**self.workflow_graph.nodes[partner_node])
    #         archive_path = section.archive_path
    #         default_sections = {
    #             'inputs': [archive_path],
    #             'outputs': [archive_path],
    #         }

    #         inouts = []
    #         for default_section in default_sections[inout_type]:
    #             flag_defaults = False
    #             if inout_type == 'outputs':
    #                 for _, _, edge2 in self.workflow_graph.out_edges(
    #                     node_source, data=True
    #                 ):
    #                     if self._check_for_defaults(inout_type, default_section,
    # edge2):
    #                         flag_defaults = True
    #                         break
    #             elif inout_type == 'inputs':
    #                 for _, _, edge2 in self.workflow_graph.in_edges(
    #                     node_dest, data=True
    #                 ):
    #                     if self._check_for_defaults(inout_type, default_section,
    # edge2):
    #                         flag_defaults = True
    #                         break
    #             if not flag_defaults:
    #                 partner_name = self.workflow_graph.nodes[partner_node].get(
    #                     'name', ''
    #                 )
    #                 path_info = self.workflow_graph.nodes[partner_node].get(
    #                     'path_info', {}
    #                 )
    #                 path_info['supersection_path'] = default_section
    #                 inouts.append(
    #                     {
    #                         'name': (
    #                             f'{inout_type[:-1]} {default_section} '
    #                             f'from {partner_name}'
    #                         ),
    #                         'path_info': path_info,
    #                         'is_default': True,
    #                     },
    #                 )
    #     # ! add more defaults here
    #     # elfif entry_type == '<default_type>':
    #     #     ...
    #     elif entry_type == 'simulation':
    #         default_sections = self.simulation_default_sections

    #         inouts = []
    #         for default_section in default_sections[inout_type]:
    #             flag_defaults = False
    #             if inout_type == 'outputs':
    #                 for _, _, edge2 in self.workflow_graph.out_edges(
    #                     node_source, data=True
    #                 ):
    #                     if self._check_for_defaults(inout_type, default_section,
    # edge2):
    #                         flag_defaults = True
    #                         break
    #             elif inout_type == 'inputs':
    #                 # don't add input defaults for edge input node
    #                 # ? Is this still relevant with the current setting of
    # partner_node?
    #                 in_tasks = [
    #                     edge[0]
    #                     for edge in self.workflow_graph.in_edges(node_dest)
    #                     if self.workflow_graph.nodes[edge[0]].get('type', '')
    #                     in ['task', 'workflow']
    #                 ]
    #                 if not in_tasks:
    #                     break
    #                 for _, _, edge2 in self.workflow_graph.in_edges(
    #                     node_dest, data=True
    #                 ):
    #                     if self._check_for_defaults(inout_type, default_section,
    # edge2):
    #                         flag_defaults = True
    #                         break
    #             if not flag_defaults:
    #                 partner_name = self.workflow_graph.nodes[partner_node].get(
    #                     'name', ''
    #                 )
    #                 inouts.append(
    #                     {
    #                         'name': (
    #                             f'{inout_type[:-1]} {default_section} '
    #                             f'from {partner_name}'
    #                         ),
    #                         'path_info': {
    #                             'section_type': default_section,
    #                             'mainfile_path':
    # self._get_mainfile_path(partner_node),
    #                         },
    #                     },
    #                 )

    #     return inouts

    def build_workflow_yaml(self) -> None:
        # register the sections and build task_elements
        # register the nodes as sections for the archive construction
        for (
            node_key,
            node_attrs,
        ) in self.workflow_graph.nodes(data=True):
            self.register_section(node_key, node_attrs)

        archive = self.generate_archive()
        archive.remove_duplicate_ios()
        archive.to_yaml(self.destination_filename)

    def generate_archive(self) -> NomadWorkflowArchive:
        archive = NomadWorkflowArchive(
            archive_section=self.archive_section, name=self.name
        )
        archive.inputs = []
        archive.outputs = []

        task_graph = self._create_task_graph()
        self._add_inputs_to_archive(archive, task_graph)
        self._add_outputs_to_archive(archive, task_graph)
        self._add_tasks_to_archive(archive, task_graph)

        return archive

    def _create_task_graph(self) -> nx.DiGraph:
        task_nodes = [
            n
            for n, attr in self.workflow_graph.nodes(data=True)
            if attr.get('type', '') in ['task', 'workflow']
        ]
        return self.workflow_graph.subgraph(task_nodes)

    def _add_inputs_to_archive(
        self, archive: NomadWorkflowArchive, task_graph: nx.DiGraph
    ) -> None:
        for node in [n for n, d in task_graph.in_degree if d == 0]:
            for edge in self.workflow_graph.in_edges(node, data=True):
                if self.workflow_graph.nodes[edge[0]].get('type', '') != 'input':
                    continue
                if self.workflow_graph.nodes[edge[0]].get('is_default', '') is True:
                    continue
                element = self.task_elements[edge[0]]
                archive.inputs.append(element)

    def _add_outputs_to_archive(
        self, archive: NomadWorkflowArchive, task_graph: nx.DiGraph
    ) -> None:
        for node in [n for n, d in task_graph.out_degree if d == 0]:
            for edge in self.workflow_graph.out_edges(node, data=True):
                if self.workflow_graph.nodes[edge[1]].get('type', '') != 'output':
                    continue
                if self.workflow_graph.nodes[edge[1]].get('is_default', '') is True:
                    continue
                element = self.task_elements[edge[1]]
                archive.outputs.append(element)

    def _add_tasks_to_archive(
        self, archive: NomadWorkflowArchive, task_graph: nx.DiGraph
    ) -> None:
        for node_key, node in task_graph.nodes(data=True):
            inputs = []
            outputs = []
            for _, _, edge in self.workflow_graph.out_edges(node_key, data=True):
                if edge.get('inputs'):
                    outputs.extend(edge.get('inputs'))
            for _, _, edge in self.workflow_graph.in_edges(node_key, data=True):
                if edge.get('outputs'):
                    inputs.extend(edge.get('outputs'))

            archive.tasks.append(
                NomadTask(
                    name=node.get('name', ''),
                    inputs=inputs,
                    outputs=outputs,
                    task_section=self.task_elements[node_key],
                )
            )

    #! Keep the below old implementation until full testing for ruff refactoring above!
    # def generate_archive(self) -> NomadWorkflowArchive:
    #     archive = NomadWorkflowArchive(
    #         archive_section=self.archive_section, name=self.name
    #     )
    #     archive.inputs = []
    #     archive.outputs = []

    #     # get the input task nodes
    #     task_nodes = [
    #         n
    #         for n, attr in self.workflow_graph.nodes(data=True)
    #         if attr.get('type', '') in ['task', 'workflow']
    #     ]
    #     # Create a subgraph with only task nodes
    #     task_graph = self.workflow_graph.subgraph(task_nodes)

    #     # select input nodes from task graph that have no incoming edges
    #     for node in [n for n, d in task_graph.in_degree if d == 0]:
    #         # get the inputs from the incoming edges of these nodes within
    #         # the full graph (should be inputs!)
    #         for edge in self.workflow_graph.in_edges(node, data=True):
    #             if self.workflow_graph.nodes[edge[0]].get('type', '') != 'input':
    #                 continue
    #             # don't add generic defaults to global in/out
    #             if self.workflow_graph.nodes[edge[0]].get('is_default', '') is True:
    #                 continue
    #             element = self.task_elements[edge[0]]
    #             archive.inputs.append(element)
    #     # select output nodes from task graph that have no outgoing edges
    #     for node in [n for n, d in task_graph.out_degree if d == 0]:
    #         # get the outputs from the outgoing edges of these nodes within
    #         # the full graph (should be outputs!)
    #         for edge in self.workflow_graph.out_edges(node, data=True):
    #             if self.workflow_graph.nodes[edge[1]].get('type', '') != 'output':
    #                 continue
    #             # don't add generic defaults to global in/out
    #             if self.workflow_graph.nodes[edge[1]].get('is_default', '') is True:
    #                 continue
    #             element = self.task_elements[edge[1]]
    #             archive.outputs.append(element)
    #     # add the tasks
    #     for node_key, node in task_graph.nodes(data=True):
    #         inputs = []
    #         outputs = []
    #         for _, _, edge in self.workflow_graph.out_edges(node_key, data=True):
    #             if edge.get('inputs'):
    #                 outputs.extend(edge.get('inputs'))
    #         for _, _, edge in self.workflow_graph.in_edges(node_key, data=True):
    #             if edge.get('outputs'):
    #                 inputs.extend(edge.get('outputs'))

    #         archive.tasks.append(
    #             NomadTask(
    #                 name=node.get('name', ''),
    #                 inputs=inputs,
    #                 outputs=outputs,
    #                 task_section=self.task_elements[node_key],
    #             )
    #         )

    #     return archive


def nodes_to_graph(node_attributes: dict[int, Any]) -> nx.DiGraph:
    """Builds a workflow graph (nx.DiGraph) from a dictionary of node attributes
    as specified below.

    Args:
        node_attributes (dict[int, Any]): _description_

    Returns:
        nx.DiGraph: _description_
    """
    if not node_attributes:
        logger.error(
            'No workflow graph or node attributes provided. Cannot build workflow.'
        )
        return None

    workflow_graph = nx.DiGraph()
    workflow_graph.add_nodes_from(node_attributes.keys())
    nx.set_node_attributes(workflow_graph, node_attributes)

    for node_key, node_attrs in list(workflow_graph.nodes(data=True)):
        _add_edges(workflow_graph, node_key, node_attrs)
        _add_global_inouts(workflow_graph, node_key, node_attrs)
        _add_task_inouts(workflow_graph, node_key, node_attrs)

    return workflow_graph


def _add_edges(workflow_graph, node_key, node_attrs):
    def set_mainfile_path(workflow_graph, edge, node_attrs) -> None:
        parent_mainfile_path = (
            workflow_graph.nodes[edge].get('path_info', '').get('mainfile_path', None)
        )
        if not node_attrs.get('path_info'):
            node_attrs['path_info'] = {'mainfile_path': parent_mainfile_path}
        else:
            node_attrs['path_info']['mainfile_path'] = node_attrs['path_info'].get(
                'mainfile_path', parent_mainfile_path
            )

    for edge in node_attrs.get('in_edge_nodes', []):
        workflow_graph.add_edge(edge, node_key)
        set_mainfile_path(workflow_graph, edge, node_attrs)
    for edge in node_attrs.get('out_edge_nodes', []):
        workflow_graph.add_edge(node_key, edge)
        set_mainfile_path(workflow_graph, edge, node_attrs)


def _add_global_inouts(workflow_graph, node_key, node_attrs):
    if node_attrs.get('type', '') == 'input':
        for edge_node in node_attrs.get('out_edge_nodes', []):
            workflow_graph.add_edge(node_key, edge_node)
    elif node_attrs.get('type', '') == 'output':
        for edge_node in node_attrs.get('in_edge_nodes', []):
            workflow_graph.add_edge(edge_node, node_key)


def _add_task_inouts(workflow_graph, node_key, node_attrs):
    inputs = node_attrs.pop('inputs', [])
    for input_ in inputs:
        edge_nodes = input_.get('out_edge_nodes', [])
        if not edge_nodes:
            edge_nodes.append(len(workflow_graph.nodes))
            workflow_graph.add_node(edge_nodes[0], type='input', **input_)

        for edge_node in edge_nodes:
            workflow_graph.add_edge(edge_node, node_key)
            if not workflow_graph.edges[edge_node, node_key].get('outputs', []):
                nx.set_edge_attributes(
                    workflow_graph, {(edge_node, node_key): {'outputs': []}}
                )
            workflow_graph.edges[edge_node, node_key]['outputs'].append(input_)

    outputs = node_attrs.pop('outputs', [])
    for output_ in outputs:
        edge_nodes = output_.get('in_edge_node', [])
        if not edge_nodes:
            edge_nodes.append(len(workflow_graph.nodes))
            workflow_graph.add_node(edge_nodes[0], type='output', **output_)

        for edge_node in edge_nodes:
            workflow_graph.add_edge(node_key, edge_node)
            if not workflow_graph.edges[node_key, edge_node].get('inputs', []):
                nx.set_edge_attributes(
                    workflow_graph, {(node_key, edge_node): {'inputs': []}}
                )
            workflow_graph.edges[node_key, edge_node]['inputs'].append(output_)


class NodeAttributes(TypedDict, total=False):
    """
    NodeAttributes represents the attributes of a node in the NOMAD workflow graph.

    Attributes:
        name (str): A free-form string describing this node, which will be used as a
                    label in the NOMAD workflow graph visualizer.

        type (Literal['input', 'output', 'workflow', 'task', 'other']):
            Specifies the type of node. Must be one of the specified options.

            - input: (meta)data taken as input for the entire workflow or a specific
                    task. For simulations, often corresponds to a section within the
                    archive (e.g., system, method).

            - output: (meta)data produced as output for the entire workflow or a
                    specific task. For simulations, often corresponds to a section
                    within the archive (e.g., calculation).

            - workflow: A node in the workflow which itself contains an internal
                    (sub)workflow, that is recognized by NOMAD. Such nodes can be
                    linked to existing workflows within NOMAD, providing
                    functionalities within NOMAD's interactive workflow graphs.

            - task: A node in the workflow which represents an individual task
                    (i.e., no underlying workflow), that is recognized by NOMAD.

            - other: A node in the workflow which represents either a (sub)workflow
                    or individual task that is not supported by NOMAD.

        entry_type (Literal['simulation']): Specifies the type of node in terms of
            tasks or workflows recognized by NOMAD. Functionally, this attribute is
            used to create default inputs and outputs that are required for properly
            creating the edge visualizations in the NOMAD GUI.

        path_info (dict): Information for generating the NOMAD archive section paths
            (i.e., connections between nodes in terms of the NOMAD MetaInfo sections).

            - upload_id (str): NOMAD PID for the upload, if exists.

            - entry_id (str): NOMAD PID for the entry, if exists.

            - mainfile_path (str): Local (relative to the native upload) path to the
                mainfile, including the mainfile name with extension.

            - supersection_path (str): Archive path to the supersection, e.g.,
                "run" or "workflow2/method".

            - supersection_index (int): The relevant index for the supersection, if it
                is a repeating subsection.

            - section_type (str): The name of the section for an input or output node,
                e.g., "system", "method", or "calculation".

            - section_index (int): The relevant index for the section, if it is a
                repeating subsection.

            - archive_path (str): Specifies the entire archive path to the section,
                e.g., "run/0/system/2".

        inputs (list[dict]): A list of input nodes to be added to the graph with
            in_edges to the parent node.

            - name (str): Will be set as the name for the input node created.

            - path_info (dict): Path information for the input node created, as
                specified for the node attributes above.

        outputs (list[dict]): A list of output nodes to be added to the graph with
            out_edges from the parent node.

            - name (str): Will be set as the name for the output node created.

            - path_info (dict): Path information for the output node created, as
                specified for the node attributes above.

        in_edge_nodes (list[int]): A list of integers specifying the node keys which
            contain in-edges to this node.

        out_edge_nodes (list[int]): A list of integers specifying the node keys which
            contain out-edges to this node.
    """

    name: str
    type: SectionType
    entry_type: EntryType
    path_info: PathInfo
    inputs: list[dict[str, Any]]
    outputs: list[dict[str, Any]]
    in_edge_nodes: list[int]
    out_edge_nodes: list[int]


def build_nomad_workflow(
    workflow_metadata: dict[str, str] = {},
    node_attributes: NodeAttributes = {},
    workflow_graph: nx.DiGraph = None,
    write_to_yaml: bool = False,
) -> nx.DiGraph:
    """_summary_

    Args:
        workflow_metadata (dict[str, str], optional): _description_. Defaults to {}.
        node_attributes (NodeAttributes, optional): _description_. Defaults to {}.
        workflow_graph (nx.DiGraph, optional): _description_. Defaults to None.
        write_to_yaml (bool, optional): _description_. Defaults to False.

    Returns:
        nx.DiGraph: _description_
    """
    destination_filename = workflow_metadata.get(
        'destination_filename', './nomad_workflow.archive.yaml'
    )
    archive_section = workflow_metadata.get('archive_section', 'workflow2')
    workflow_name = workflow_metadata.get('workflow_name', '')
    workflow = NomadWorkflow(
        destination_filename=destination_filename,
        archive_section=archive_section,
        name=workflow_name,
        node_attributes=node_attributes,
        workflow_graph=workflow_graph,
    )
    if write_to_yaml:
        workflow.build_workflow_yaml()

    return workflow.workflow_graph


# TODO the input from in edge task nodes are automatically added to the global inputs...
# TODO but not vice versa, the reverse should be done...
# TODO also prevent that the same ios are added 2x
# TODO I need to check that the defaults are generated properly when you have multiple
# input or output task nodes.
# TODO we need to fix the default inputs, so that system[-1] is not added, and instead
# either the global input or possibly system[0] only
# TODO -1 notation doesn't work for run for connections!!
# TODO test this code on a number of already existing examples
# TODO create docs with some examples for dict and graph input types
# TODO add to readme/docs that this is not currently using NOMAD, but could be linked
# later?
# TODO add some text to the test notebooks
# TODO change the rest of the functions to pydantic -- not sure if I really want to
# tackle this now



import datetime as dt
import logging
from typing import Optional, TypedDict

from marshmallow import EXCLUDE, Schema, pre_load
from marshmallow_dataclass import class_schema, dataclass

from nomad_utility_workflows.utils.core import (
    RequestOptions,
    delete_nomad_request,
    get_nomad_request,
    get_nomad_url,
    get_nomad_url_name,
    post_nomad_request,
)
from nomad_utility_workflows.utils.users import NomadUser, get_user_by_id

logger = logging.getLogger(__name__)


class NomadDatasetSchema(Schema):
    @pre_load
    def convert_users(self, data, **kwargs):
        data['user'] = get_user_by_id(user_id=data['user_id']).as_dict()
        del data['user_id']
        return data


class DatasetParams(TypedDict, total=False):
    dataset_id: str
    dataset_name: str
    user_id: str
    page_size: int
    max_datasets: int


default_dataset_params = {
    'dataset_id': None,
    'dataset_name': None,
    'user_id': None,
    'page_size': 10,
    'max_datasets': 50,
}


@dataclass(frozen=True)
class NomadDataset:
    class Meta:
        unknown = EXCLUDE

    dataset_id: str
    dataset_create_time: dt.datetime
    dataset_name: str
    dataset_type: Optional[str] = None
    dataset_modified_time: Optional[dt.datetime] = None
    user: Optional[NomadUser] = None
    doi: Optional[str] = None
    pid: Optional[int] = None
    m_annotations: Optional[dict] = None


def retrieve_datasets(
    dataset_params: DatasetParams = default_dataset_params.copy(),
    url: str = None,
) -> list[NomadDataset]:
    parameters = []
    max_datasets = dataset_params.pop(
        'max_datasets', default_dataset_params['max_datasets']
    )
    for key, value in dataset_params.items():
        parameters.append(f'{key}={value}')
    url = get_nomad_url(url)
    section = '/datasets/'
    if len(parameters) > 0:
        section += f'?{parameters[0]}'
    for i in range(1, len(parameters)):
        section += f'&{parameters[i]}'
    headers = {'Accept': 'application/json'}
    nomad_entry_schema = class_schema(NomadDataset, base_schema=NomadDatasetSchema)
    datasets = []
    page_after_value = None
    while (max_datasets > 0 and len(datasets) <= max_datasets) or (max_datasets < 0):
        section = (
            f'{section}&page_after_value={page_after_value}'
            if page_after_value
            else section
        )
        response = get_nomad_request(
            RequestOptions(section=section, headers=headers, url=url)
        )
        if len(response['data']) == 0:
            break
        datasets.extend([nomad_entry_schema().load(d) for d in response['data']])
        if response['pagination']['page'] == response['pagination']['total']:
            break
        page_after_value = response['pagination'].get('next_page_after_value')
    return datasets


def get_dataset_by_id(dataset_id: str, url: str = None) -> NomadDataset:
    datasets = retrieve_datasets(DatasetParams(dataset_id=dataset_id), url=url)
    if len(datasets) != 1:
        raise ValueError(f'Problem retrieving dataset {dataset_id}: {datasets}')
    return datasets[0]


def create_dataset(dataset_name: str, url: str = None, timeout_in_sec: int = 10) -> str:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('creating dataset name %s on %s server', dataset_name, url_name)
    json_dict = {'dataset_name': dataset_name}
    response = post_nomad_request(
        RequestOptions(
            section='/datasets/',
            with_authentication=True,
            url=url,
            timeout_in_sec=timeout_in_sec,
        ),
        json_dict=json_dict,
    )
    return response.get('dataset_id')


def delete_dataset(dataset_id: str, url: str = None, timeout_in_sec: int = 10) -> None:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('deleting dataset %s on %s server', dataset_id, url_name)
    response = delete_nomad_request(
        RequestOptions(
            section=f'/datasets/{dataset_id}',
            with_authentication=True,
            url=url,
            timeout_in_sec=timeout_in_sec,
        )
    )
    if response.get('dataset_id'):
        logger.info('successfully deleted dataset %s', dataset_id)
    else:
        logger.error('no dataset deleted')



import datetime as dt
import logging
from typing import Any, Optional, TypedDict

from cachetools.func import ttl_cache
from marshmallow import EXCLUDE, Schema, pre_load
from marshmallow_dataclass import class_schema, dataclass

from nomad_utility_workflows.utils.core import (
    RequestOptions,
    delete_nomad_request,
    get_nomad_base_url,
    get_nomad_request,
    get_nomad_url,
    get_nomad_url_name,
    post_nomad_request,
)
from nomad_utility_workflows.utils.users import NomadUser, get_user_by_id

logger = logging.getLogger(__name__)


class NomadUploadSchema(Schema):
    @pre_load
    def convert_users(self, data, **kwargs):
        data['main_author'] = get_user_by_id(user_id=data['main_author']).as_dict()
        data['writers'] = [get_user_by_id(user_id=w).as_dict() for w in data['writers']]
        data['reviewers'] = [
            get_user_by_id(user_id=r).as_dict() for r in data['reviewers']
        ]
        data['viewers'] = [get_user_by_id(user_id=v).as_dict() for v in data['viewers']]
        return data


class UploadMetadata(TypedDict, total=False):
    class Meta:
        unknown = EXCLUDE

    upload_name: str
    references: list[str]
    dataset_id: str
    embargo_length: float
    coauthors_ids: list[str]
    comment: str


@dataclass(frozen=True)
class NomadUpload:
    class Meta:
        unknown = EXCLUDE

    upload_id: str
    upload_create_time: dt.datetime
    main_author: NomadUser
    process_running: bool
    current_process: str
    process_status: str
    last_status_message: str
    errors: list[Any]
    warnings: list[Any]
    coauthors: list[str]
    coauthor_groups: list[Any]
    reviewers: list[NomadUser]
    reviewer_groups: list[Any]
    writers: list[NomadUser]
    writer_groups: list[Any]
    viewers: list[NomadUser]
    viewer_groups: list[Any]
    published: bool
    published_to: list[Any]
    with_embargo: bool
    embargo_length: float
    license: str
    entries: int
    n_entries: Optional[int] = None
    upload_files_server_path: Optional[str] = None
    publish_time: Optional[dt.datetime] = None
    references: Optional[list[str]] = None
    datasets: Optional[list[str]] = None
    external_db: Optional[str] = None
    upload_name: Optional[str] = None
    comment: Optional[str] = None
    url: Optional[str] = None
    complete_time: Optional[dt.datetime] = None

    @property
    def base_url(self) -> Optional[str]:
        url = get_nomad_url(self.url)
        return get_nomad_base_url(url)

    @property
    def nomad_gui_url(self) -> str:
        if self.upload_id is None:
            raise ValueError(f"missing attributes 'upload_id' for entry {self}")
        return f'{self.base_url}/gui/user/uploads/upload/id/{self.upload_id}'


@ttl_cache(maxsize=128, ttl=180)
def get_all_my_uploads(url: str = None, timeout_in_sec: int = 10) -> list[NomadUpload]:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving all uploads on %s server', url_name)
    response = get_nomad_request(
        RequestOptions(
            section='/uploads',
            url=url,
            with_authentication=True,
            timeout_in_sec=timeout_in_sec,
        )
    )
    upload_class_schema = class_schema(NomadUpload, base_schema=NomadUploadSchema)
    return [upload_class_schema().load({**r, 'url': url}) for r in response['data']]


def get_upload_by_id(
    upload_id: str, url: str = None, timeout_in_sec: int = 10
) -> NomadUpload:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving upload %s on %s server', upload_id, url_name)
    response = get_nomad_request(
        RequestOptions(
            section=f'/uploads/{upload_id}',
            url=url,
            with_authentication=True,
            timeout_in_sec=timeout_in_sec,
        )
    )
    upload_class_schema = class_schema(NomadUpload, base_schema=NomadUploadSchema)
    return upload_class_schema().load({**response['data'], 'url': url})


def delete_upload(
    upload_id: str, url: str = None, timeout_in_sec: int = 10
) -> NomadUpload:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info(f'deleting upload {upload_id} on {url_name} server')
    response = delete_nomad_request(
        RequestOptions(
            section=f'/uploads/{upload_id}',
            with_authentication=True,
            url=url,
            timeout_in_sec=timeout_in_sec,
        )
    )
    upload_class_schema = class_schema(NomadUpload, base_schema=NomadUploadSchema)
    return upload_class_schema().load({**response['data'], 'url': url})


def upload_files_to_nomad(
    filename: str, url: str = None, timeout_in_sec: int = 30
) -> str:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('uploading file %s on %s server', filename, url_name)
    with open(filename, 'rb') as f:
        response = post_nomad_request(
            RequestOptions(
                section='/uploads',
                with_authentication=True,
                url=url,
                timeout_in_sec=timeout_in_sec,
            ),
            data=f,
        )
    upload_id = response.get('upload_id')
    if upload_id:
        logger.info('successful upload to %s', upload_id)
        return upload_id
    else:
        logger.error('could not upload %s. Response %s', filename, response)


def publish_upload(upload_id: str, url: str = None, timeout_in_sec: int = 10) -> dict:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('publishing upload %s on %s server', upload_id, url_name)
    response = post_nomad_request(
        RequestOptions(
            section=f'/uploads/{upload_id}/action/publish',
            with_authentication=True,
            url=url,
            timeout_in_sec=timeout_in_sec,
        )
    )
    return response


def edit_upload_metadata(
    upload_id: str,
    upload_metadata: UploadMetadata = {},
    url: str = None,
    timeout_in_sec: int = 10,
) -> dict:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('editing the metadata for upload %s on %s server', upload_id, url_name)
    metadata = {'metadata': {}}
    if 'dataset_id' in upload_metadata.keys():
        upload_metadata['datasets'] = upload_metadata.pop('dataset_id')
    for key, value in upload_metadata.items():
        metadata['metadata'][key] = value
    response = post_nomad_request(
        RequestOptions(
            section=f'/uploads/{upload_id}/edit',
            url=url,
            with_authentication=True,
            timeout_in_sec=timeout_in_sec,
        ),
        json_dict=metadata,
    )
    return response



import json
import logging
from typing import Any, TypedDict

import requests
from cachetools.func import ttl_cache
from decouple import config as environ

logger = logging.getLogger(__name__)

NOMAD_USERNAME = environ('NOMAD_USERNAME')
NOMAD_PASSWORD = environ('NOMAD_PASSWORD')
NOMAD_PROD_URL = 'https://nomad-lab.eu/prod/v1/api/v1'
NOMAD_STAGING_URL = 'https://nomad-lab.eu/prod/v1/staging/api/v1'
NOMAD_TEST_URL = 'https://nomad-lab.eu/prod/v1/test/api/v1'
STATUS_CODE = 200
TIMEOUT_IN_SEC = 60


@ttl_cache(maxsize=128, ttl=180)
def get_authentication_token(
    url: str = None,
    username: str = NOMAD_USERNAME,
    password: str = NOMAD_PASSWORD,
    timeout_in_sec: int = TIMEOUT_IN_SEC,
) -> str:
    """_summary_

    Args:
        url (str, optional): The NOMAD API URL. Defaults to None.
        username (str, optional): _description_. Defaults to NOMAD_USERNAME.
        password (str, optional): _description_. Defaults to NOMAD_PASSWORD.
        timeout_in_sec (int, optional): _description_. Defaults to TIMEOUT_IN_SEC.

    Raises:
        ValueError: _description_

    Returns:
        str: _description_
    """
    url = get_nomad_url(url)
    logger.info('Requesting authentication token @ %s', url)
    response = requests.get(
        url + '/auth/token',
        params={'username': username, 'password': password},
        timeout=timeout_in_sec,
    )
    if not response.status_code == STATUS_CODE:
        raise ValueError(f'Unexpected response {response.json()}')
    return response.json().get('access_token')


class RequestOptions(TypedDict, total=False):
    """_summary_

    Args:
        TypedDict (_type_): _description_
        total (bool, optional): _description_. Defaults to False.
    """

    section: str
    url: str = None
    timeout_in_sec: int = TIMEOUT_IN_SEC
    headers: dict = None
    with_authentication: bool = False


default_request_options = {
    'section': None,
    'url': None,
    'timeout_in_sec': TIMEOUT_IN_SEC,
    'headers': None,
    'with_authentication': False,
}


def get_nomad_request(
    request_options: RequestOptions = default_request_options.copy(),
    return_json: bool = True,
    accept_field: str = 'application/json',
) -> Any:
    """_summary_

    Args:
        section (str): _description_
        url (str, optional): The NOMAD API URL. Defaults to None.
        timeout_in_sec (int, optional): _description_. Defaults to TIMEOUT_IN_SEC.
        headers (dict, optional): _description_. Defaults to None.
        with_authentication (bool, optional): _description_. Defaults to False.
        return_json (bool, optional): _description_. Defaults to True.
        accept_field (str, optional): _description_. Defaults to "application/json".

    Raises:
        ValueError: _description_

    Returns:
        Any: _description_
    """
    section = request_options.get('section')
    url = request_options.get('url')
    timeout_in_sec = request_options.get('timeout_in_sec')
    headers = request_options.get('headers')
    with_authentication = request_options.get('with_authentication')

    url_base = get_nomad_url(url)
    url = url_base + f"{'/' if section[0] != '/' else ''}{section}"
    logger.info('Sending get request @ %s', url)
    if headers is None:
        headers = {}
    if with_authentication:
        token = get_authentication_token(url=url_base)
        headers |= {
            'Authorization': f'Bearer {token}',
            'Accept': accept_field,
        }
    response = requests.get(url, headers=headers, timeout=timeout_in_sec)
    if not response.status_code == STATUS_CODE:
        raise ValueError(f'Unexpected response {response.json()}')
    if return_json:
        return response.json()
    return response.content


def get_nomad_url_name(url: str) -> str:
    """_summary_

    Args:
        url (str): _description_

    Returns:
        str: _description_
    """
    try:
        return url.split('/')[-3]
    except IndexError:
        return 'unknown'


def get_nomad_url(url: str) -> str:
    """
    Get the Nomad URL based on the given URL. If no URL is given, the NOMAD Test URL
    is returned by default. If the given URL is "prod", "staging" or "test",
    the corresponding Nomad URL will be returned. Otherwise, the given URL will be
    returned.
    """
    if url is None:
        return NOMAD_TEST_URL
    elif url == 'prod':
        return NOMAD_PROD_URL
    elif url == 'staging':
        return NOMAD_STAGING_URL
    elif url == 'test':
        return NOMAD_TEST_URL

    if not url.endswith('/api/v1'):
        logger.warning(
            'The given URL: %s does not appear to be a valid NOMAD API URL, '
            'i.e., ending with /api/v1.',
            url,
        )

    return url


def get_nomad_base_url(url: str) -> str:
    """_summary_

    Args:
        url (str): _description_

    Returns:
        str: _description_
    """
    return (get_nomad_url(url)).removesuffix('/api/v1')


def post_nomad_request(
    request_options: RequestOptions = default_request_options.copy(),
    data: Any = None,
    json_dict: dict = None,
) -> json:
    """_summary_

    Args:
        request_options (RequestOptions, optional): _description_.
            Defaults to default_request_options.copy().
        data (Any, optional): _description_. Defaults to None.
        json_dict (dict, optional): _description_. Defaults to None.

    Raises:
        ValueError: _description_

    Returns:
        json: _description_
    """
    section = request_options.get('section')
    url = request_options.get('url')
    timeout_in_sec = request_options.get('timeout_in_sec')
    headers = request_options.get('headers')
    with_authentication = request_options.get('with_authentication')

    if headers is None:
        headers = {}
    if with_authentication:
        token = get_authentication_token(url=url)
        headers |= {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
        }
    if data is None:
        data = {}
    if json_dict is None:
        json_dict = {}
    url = get_nomad_url(url)
    url += f"{'/' if section[0] != '/' else ''}{section}"
    logger.info('Sending post request @ %s', url)
    response = requests.post(
        url, headers=headers, json=json_dict, data=data, timeout=timeout_in_sec
    )
    if not response.status_code == STATUS_CODE:
        raise ValueError(f'Unexpected response {response.json()}')
    return response.json()


def delete_nomad_request(
    request_options: RequestOptions = default_request_options.copy(),
) -> json:
    """_summary_

    Args:
        request_options (RequestOptions, optional): _description_.
            Defaults to default_request_options.copy().

    Raises:
        ValueError: _description_

    Returns:
        json: _description_
    """
    section = request_options.get('section')
    url = request_options.get('url')
    timeout_in_sec = request_options.get('timeout_in_sec')
    headers = request_options.get('headers')
    with_authentication = request_options.get('with_authentication')

    if headers is None:
        headers = {}
    if with_authentication:
        token = get_authentication_token(url=url)
        headers |= {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
        }
    url = get_nomad_url(url)
    url += f"{'/' if section[0] != '/' else ''}{section}"
    logger.info('Sending delete request @ %s', url)
    response = requests.delete(url, headers=headers, timeout=timeout_in_sec)
    if not response.status_code == STATUS_CODE:
        raise ValueError(f'Unexpected response {response.json()}')
    return response.json()



import datetime as dt
import logging
from dataclasses import asdict, field
from typing import Optional

from cachetools.func import ttl_cache
from marshmallow import EXCLUDE
from marshmallow_dataclass import class_schema, dataclass

from nomad_utility_workflows.utils.core import (
    RequestOptions,
    get_nomad_request,
    get_nomad_url,
    get_nomad_url_name,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class NomadUser:
    class Meta:
        unknown = EXCLUDE

    user_id: str = field(repr=False)
    name: str
    first_name: str = field(repr=False)
    last_name: str = field(repr=False)
    username: str = field(repr=False)
    affiliation: str = field(repr=False)
    affiliation_address: str = field(repr=False)
    email: Optional[str] = field(repr=False, default=None)
    is_oasis_admin: Optional[bool] = field(repr=False, default=None)
    is_admin: Optional[bool] = field(repr=False, default=None)
    repo_user_id: Optional[str] = field(repr=False, default=None)
    created: Optional[dt.datetime] = field(repr=False, default=None)

    def as_dict(self) -> dict:
        return asdict(self)


@ttl_cache(maxsize=128, ttl=180)
def search_users_by_name(
    user_name: str, url: str = None, timeout_in_sec: int = 10
) -> NomadUser:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving user %s on %s server', user_name, url_name)
    response = get_nomad_request(
        RequestOptions(
            section=f'/users?prefix={user_name}', timeout_in_sec=timeout_in_sec, url=url
        )
    ).get('data', [])
    return [class_schema(NomadUser)().load(user) for user in response]


@ttl_cache(maxsize=128, ttl=180)
def get_user_by_id(
    user_id: str, url: str = None, timeout_in_sec: int = 10
) -> NomadUser:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving user %s on %s server', user_id, url_name)
    response = get_nomad_request(
        RequestOptions(
            section=f'/users/{user_id}', timeout_in_sec=timeout_in_sec, url=url
        )
    )
    user_schema = class_schema(NomadUser)
    return user_schema().load(response)


@ttl_cache(maxsize=128, ttl=180)
def who_am_i(url: str = None, timeout_in_sec: int = 10) -> NomadUser:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving self user info on %s server', url_name)
    response = get_nomad_request(
        RequestOptions(
            section='/users/me',
            with_authentication=True,
            timeout_in_sec=timeout_in_sec,
            url=url,
        )
    )
    user_schema = class_schema(NomadUser)
    return user_schema().load(response)



import datetime as dt
import json
import logging
from collections.abc import ByteString
from dataclasses import field
from typing import Any, Optional, TypedDict

from cachetools.func import ttl_cache
from marshmallow import EXCLUDE, Schema, pre_load
from marshmallow_dataclass import class_schema, dataclass

from nomad_utility_workflows.utils.core import (
    RequestOptions,
    get_nomad_base_url,
    get_nomad_request,
    get_nomad_url,
    get_nomad_url_name,
    post_nomad_request,
)

# from signac.job import Job
# from martignac import config
from nomad_utility_workflows.utils.datasets import NomadDataset
from nomad_utility_workflows.utils.uploads import get_all_my_uploads
from nomad_utility_workflows.utils.users import NomadUser, get_user_by_id

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class NomadSectionDefinition:
    used_directly: bool
    definition_id: str
    definition_qualified_name: str


class NomadEntrySchema(Schema):
    class Meta:
        unknown = EXCLUDE

    @pre_load
    def convert_users(self, data, **kwargs):
        data['main_author'] = get_user_by_id(
            user_id=data['main_author']['user_id']
        ).as_dict()
        data['writers'] = [
            get_user_by_id(user_id=w['user_id']).as_dict() for w in data['writers']
        ]
        data['authors'] = [
            get_user_by_id(user_id=a['user_id']).as_dict() for a in data['authors']
        ]
        data['viewers'] = [
            get_user_by_id(user_id=v['user_id']).as_dict() for v in data['viewers']
        ]
        return data


class QueryParams(TypedDict, total=False):
    worfklow_name: str
    program_name: str
    dataset_id: str
    origin: str
    page_size: int
    max_entries: int


default_query_params = {
    'worfklow_name': None,
    'program_name': None,
    'dataset_id': None,
    'origin': None,
    'page_size': 10,
    'max_entries': 50,
}


# ? Should we just make these all optional to avoid the get functions breaking?
@dataclass(frozen=True)
class NomadEntry:
    class Meta:
        unknown = EXCLUDE

    entry_id: str
    upload_id: str
    references: list[str]
    origin: str
    quantities: list[str] = field(repr=False)
    datasets: list[NomadDataset] = field(repr=False)
    n_quantities: int
    nomad_version: str
    upload_create_time: dt.datetime
    nomad_commit: str
    section_defs: list[NomadSectionDefinition] = field(repr=False)
    processing_errors: list[Any]
    results: dict = field(repr=False)
    entry_name: str
    last_processing_time: dt.datetime
    parser_name: str
    calc_id: str
    published: bool
    writers: list[NomadUser]
    sections: list[str] = field(repr=False)
    processed: bool
    mainfile: str
    main_author: NomadUser
    viewers: list[NomadUser] = field(repr=False)
    entry_create_time: dt.datetime
    with_embargo: bool
    files: list[str] = field(repr=False)
    entry_type: Optional[str]
    authors: list[NomadUser] = field(repr=False)
    license: str
    domain: Optional[str] = None
    optimade: Optional[dict] = field(repr=False, default=None)
    comment: Optional[str] = None
    upload_name: Optional[str] = None
    viewer_groups: Optional[list[Any]] = field(repr=False, default=None)
    writer_groups: Optional[list[Any]] = field(repr=False, default=None)
    text_search_contents: Optional[list[str]] = None
    publish_time: Optional[dt.datetime] = None
    entry_references: Optional[list[dict]] = None
    url: Optional[str] = None

    @property
    def base_url(self) -> Optional[str]:
        url = get_nomad_url(self.url)
        return get_nomad_base_url(url)

    @property
    def nomad_gui_url(self) -> str:
        if self.upload_id is None or self.entry_id is None:
            raise ValueError(
                f"missing attributes 'upload_id' or 'entry_id' for entry {self}"
            )
        return (
            f'{self.base_url}/gui/user/uploads/upload/id/{self.upload_id}/entry/id/'
            f'{self.entry_id}'
        )

    @property
    def job_id(self) -> Optional[str]:
        return self._comment_dict.get('job_id', None)

    @property
    def workflow_name(self) -> Optional[str]:
        return self._comment_dict.get('workflow_name', None)

    @property
    def state_point(self) -> dict:
        return self._comment_dict.get('state_point', {})

    @property
    def mdp_files(self) -> Optional[str]:
        return self._comment_dict.get('mdp_files', None)

    @property
    def _comment_dict(self) -> dict:
        return json.loads(self.comment or '{}')


@ttl_cache(maxsize=128, ttl=180)
def get_entry_by_id(
    entry_id: str,
    url: str = None,
    with_authentication: bool = False,
    timeout_in_sec: int = 10,
) -> NomadEntry:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving entry %s on %s server', entry_id, url_name)
    response = get_nomad_request(
        RequestOptions(
            section=f'/entries/{entry_id}',
            with_authentication=with_authentication,
            url=url,
            timeout_in_sec=timeout_in_sec,
        )
    )
    nomad_entry_schema = class_schema(NomadEntry, base_schema=NomadEntrySchema)
    return nomad_entry_schema().load({**response['data'], 'url': url})


@ttl_cache(maxsize=128, ttl=180)
def get_entries_of_upload(
    upload_id: str,
    url: str = None,
    with_authentication: bool = False,
    timeout_in_sec: int = 10,
) -> list[NomadEntry]:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving entries for upload %s on %s server', upload_id, url_name)
    response = get_nomad_request(
        RequestOptions(
            section=f'/uploads/{upload_id}/entries',
            with_authentication=with_authentication,
            url=url,
            timeout_in_sec=timeout_in_sec,
        )
    )
    nomad_entry_schema = class_schema(NomadEntry, base_schema=NomadEntrySchema)
    return [
        nomad_entry_schema().load({**r['entry_metadata'], 'url': url})
        for r in response['data']
    ]


def get_entries_of_my_uploads(
    url: str = None, timeout_in_sec: int = 10
) -> list[NomadEntry]:
    return [
        upload_entry
        for u in get_all_my_uploads(url=url, timeout_in_sec=timeout_in_sec)
        for upload_entry in get_entries_of_upload(
            u.upload_id, with_authentication=True, url=url
        )
    ]


# @ttl_cache(maxsize=128, ttl=180)
# ! Had to remove caching because of the use of dict as input
# ! which was required to reduce the number of inputs for ruff
def query_entries(
    query_params: QueryParams = default_query_params.copy(),
    url: str = None,
) -> list[NomadEntry]:
    json_dict = {
        'query': {},
        'pagination': {
            'page_size': query_params.pop(
                'page_size', default_query_params['page_size']
            )
        },
        'required': {'include': ['entry_id']},
    }
    entries = []
    max_entries = query_params.pop('max_entries', default_query_params['max_entries'])
    while (max_entries > 0 and len(entries) <= max_entries) or (max_entries < 0):
        if 'dataset_id' in query_params.keys():
            json_dict['query']['datasets'] = {'dataset_id': query_params['dataset_id']}
        if 'worfklow_name' in query_params.keys():
            json_dict['query']['results.method'] = {
                'workflow_name': query_params['worfklow_name']
            }
        if 'program_name' in query_params.keys():
            json_dict['query']['results.method'] = {
                'simulation': {'program_name': query_params['program_name']}
            }
        if 'origin' in query_params.keys():
            json_dict['query']['origin'] = query_params['origin']

        query = post_nomad_request(
            RequestOptions(section='/entries/query', url=url), json_dict=json_dict
        )
        entries.extend([q['entry_id'] for q in query['data']])
        next_page_after_value = query['pagination'].get('next_page_after_value', None)
        if next_page_after_value:
            json_dict['pagination']['page_after_value'] = next_page_after_value
        else:
            break
    if max_entries > 0:
        entries = entries[:max_entries]
    return [get_entry_by_id(e, url=url) for e in entries]


def download_entry_raw_data_by_id(
    entry_id: str,
    url: str = None,
    timeout_in_sec: int = 10,
    with_authentication: bool = False,
    zip_file_name: str = None,
) -> ByteString:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)
    logger.info('retrieving raw data of entry ID %s on %s server', entry_id, url_name)
    response = get_nomad_request(
        RequestOptions(
            section=f'/entries/{entry_id}/raw?compress=true',
            with_authentication=with_authentication,
            url=url,
            timeout_in_sec=timeout_in_sec,
        ),
        return_json=False,
        accept_field='application/zip',
    )

    if zip_file_name is not None:
        with open(zip_file_name, 'wb') as f:
            f.write(bytes(response))

    return response


def download_entry_by_id(
    entry_id: str,
    url: str = None,
    timeout_in_sec: int = 10,
    with_authentication: bool = False,
    zip_file_name: str = None,
) -> dict:
    url = get_nomad_url(url)
    url_name = get_nomad_url_name(url)

    logger.info('retrieving data of entry ID %s on %s server', entry_id, url_name)
    response = get_nomad_request(
        RequestOptions(
            section=f'/entries/{entry_id}/archive/download',
            with_authentication=with_authentication,
            url=url,
            timeout_in_sec=timeout_in_sec,
        )
    )

    if zip_file_name is not None:
        __ = download_entry_raw_data_by_id(
            entry_id=entry_id,
            url=url,
            timeout_in_sec=timeout_in_sec,
            with_authentication=with_authentication,
            zip_file_name=zip_file_name,
        )

    return response
